	-----------------------------------------------------------------------------------------------
-- Client Lua Script for MapPingMarker
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
require "GameLib"
require "GroupLib"
require "ICCommLib" 

-----------------------------------------------------------------------------------------------
-- MapPingMarker Module Definition
-----------------------------------------------------------------------------------------------
local MapPingMarker = {}

--TODO: Cleanup, Better direction handling

--[[
	ChangeLog:
		1.2:
			-Pings now work while on a mount or taxi
			-Improved Direction Tracker to work correctly
		1.1:
			-Added TrackMaster support
			-Sounds from buttons in options menu
			-Moved default display position to be centered and not overlapping things
			-Added tooltips for all the options in the options menu
		1.0:
			-Initial Release	
]]

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------

function MapPingMarker:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
	self.tOptions = 
	{	
		bMarkOwnPings = true,
		bDistanceDisplayLocked = true,
		bEnableDebugMode = false,
		iProximityThreshold = 3,
		tDistanceWindowPositioning = { left = 845, top = 570, right = 1100, bottom = 690 },
		bEnableTrackMasterSupport = true
	}
	self.bTrackMasterInstalled = false	

	self.strComChannelName = "MPMComChannel"
	self.mComChannel = nil
	self.tCurrentTrackedPosition = nil
    return o
end

function MapPingMarker:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = "MapPingMarker"

    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, nil)
end

function MapPingMarker:OnLoad()
	self.xmlDoc = XmlDoc.CreateFromFile("MapPingMarker.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	
	self.mComChannel = ICCommLib.JoinChannel(self.strComChannelName, "onPacketRecieved", self)
end

function MapPingMarker:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.optionsMenu = Apollo.LoadForm(self.xmlDoc, "OptionsMenu", nil, self)
		self.distanceDisplay = Apollo.LoadForm(self.xmlDoc, "DistanceDisplay", nil, self)
		
	    self.optionsMenu:Show(false, true)	
		self:InitOptions()

		self.distanceDisplay:Show(false, true)	
		self:setDistanceDisplayPositioning(self.distanceDisplay:GetAnchorOffsets())
		
		self.bTrackMasterInstalled = Apollo.GetAddon("TrackMaster")

		Apollo.RegisterSlashCommand("mpm", "ShowOptionsMenu", self)
		
		Apollo.RegisterEventHandler("ZoneMapPing", "OnMapPing", self)
		
		Apollo.CreateTimer("UpdateDistanceTimer", .5, true)
		Apollo.StopTimer("UpdateDistanceTimer")
		Apollo.RegisterTimerHandler("UpdateDistanceTimer", "onDistanceTimer", self)
		
	end
end

function MapPingMarker:OnSave(eLevel)
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
        return nil
    end
		
	local saveData = {}
	
	saveData["Options"] = self.tOptions	
	
	return saveData
end

function MapPingMarker:OnRestore(eLevel, tData)
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
        return nil
    end

	if(tData["Options"] ~= nil) then
		self.tOptions = tData["Options"]
	end
end

function MapPingMarker:InitOptions()
		self.optionsMenu:FindChild("Options"):FindChild("MarkOwnMapPings"):SetCheck(self.tOptions.bMarkOwnPings)
		self.optionsMenu:FindChild("Options"):FindChild("DisplayLock"):SetCheck(self.tOptions.bDistanceDisplayLocked)
		self.optionsMenu:FindChild("Options"):FindChild("EnableTrackMasterSupport"):SetCheck(self.tOptions.bEnableTrackMasterSupport)
		self.optionsMenu:FindChild("Options"):FindChild("EnableDebugMode"):SetCheck(self.tOptions.bEnableDebugMode)
		
		self.optionsMenu:FindChild("ProximityThreshold"):SetValue(self.tOptions.iProximityThreshold)
		self.optionsMenu:FindChild("ProximityThreshold"):FindChild("ProxThresh Value"):SetText(tostring(self.tOptions.iProximityThreshold .. " Meters"))
		
		left = self.tOptions.tDistanceWindowPositioning.left
		top = self.tOptions.tDistanceWindowPositioning.top 
		right = self.tOptions.tDistanceWindowPositioning.right 
		bottom = self.tOptions.tDistanceWindowPositioning.bottom 

		self.distanceDisplay:SetAnchorOffsets(left, top, right, bottom)
end


function MapPingMarker:restoreDefaults()
	
	self.tOptions = 
	{	
		bMarkOwnPings = true,
		bDistanceDisplayLocked = true,
		bEnableDebugMode = false,
		iProximityThreshold = 3,
		tDistanceWindowPositioning = { left = 845, top = 570, right = 1100, bottom = 690 },
		bEnableTrackMasterSupport = true

	}
		
	self:InitOptions()	
end

-----------------------------------------------------------------------------------------------
-- MapPingMarker General Functions
-----------------------------------------------------------------------------------------------

function MapPingMarker:OnMapPing(idUnit, tPos)
	local tPlayerIds = {
		iPlayerId = self:getUnitId(GameLib.GetPlayerUnit()),
		iMountId = self:getUnitId(GameLib.GetPlayerMountUnit()),
		iTaxiId = self:getUnitId(GameLib.GetPlayerTaxiUnit())
	}

	for _, id in pairs(tPlayerIds) do
		if (id ~= nil and idUnit == id) then
			self:sendPacket(self:buildPacket(tPos))
			break
		end	
	end
end

function MapPingMarker:onDistanceTimer()
	if(self.currentTrackedPosition ~= nil) then
		distance = math.floor(self:getDistanceTo(self.currentTrackedPosition))
		if(distance > self.tOptions.iProximityThreshold) then
			self.distanceDisplay:FindChild("DistanceText"):SetText(distance .. "m")
			self:updateDirectionSlider()
		else
			self:resetMarker()
		end
	end
end

function MapPingMarker:updateDirectionSlider()
	--[[
	New System to update the slider correctly, needs testing
	]]
	local tPlayerDirection = GameLib.GetPlayerUnit():GetFacing()
	local tPlayerPosition = GameLib.GetPlayerUnit():GetPosition()
	
	local iPlayerFacingAngle = math.deg(math.atan2(tPlayerDirection.z, tPlayerDirection.x))
	local iPlayerAngleToPing = math.deg(math.atan2(self.currentTrackedPosition.z - tPlayerPosition.z, self.currentTrackedPosition.x - tPlayerPosition.x))
	
	--Set the slider so that the middle is equal to the angle from the player to the ping
	local iNormalMin = -180
	local iNormalMax = 180
	self.distanceDisplay:FindChild("DirectionSlider"):SetMinMax(iNormalMin + iPlayerAngleToPing, iNormalMax + iPlayerAngleToPing)

	local tTriangle = self:buildTriangle(Vector2.New(tPlayerPosition.x, tPlayerPosition.z), Vector2.New(0,0), Vector2.New(self.currentTrackedPosition.x, self.currentTrackedPosition.z))
	
	local iAngleNeededToTurn = 0
	if(iPlayerFacingAngle >= 0) then
		iAngleNeededToTurn = iPlayerFacingAngle - tTriangle.iAngleA 
	else 
		iAngleNeededToTurn = iPlayerFacingAngle + tTriangle.iAngleA 
	end
	
	self.distanceDisplay:FindChild("DirectionSlider"):SetValue(iAngleNeededToTurn)
end

function MapPingMarker:setMarker(tPosition)
	self.currentTrackedPosition = tPosition
	Apollo.StartTimer("UpdateDistanceTimer")

	self.distanceDisplay:Show(true, false)
	
	if(self.tOptions.bEnableTrackMasterSupport and self.bTrackMasterInstalled) then
		Apollo.GetAddon("TrackMaster"):SetTarget(Vector3.New(tPosition.x, tPosition.y, tPosition.z), self.tOptions.iProximityThreshold, nil)
	end
end

function MapPingMarker:resetMarker()
	self.currentTrackedPosition = nil
	Apollo.StopTimer("UpdateDistanceTimer")
	
	self.distanceDisplay:Show(false, false)
	
	if(self.tOptions.bEnableTrackMasterSupport and self.bTrackMasterInstalled) then
		Apollo.GetAddon("TrackMaster"):SetTarget(nil) 
	end

end

-----------------------------------------------------------------------------------------------
-- MapPingMarker Communication Functions
-----------------------------------------------------------------------------------------------

function MapPingMarker:buildPacket(tPosition)
	local tPacket = {}
	tPacket.strPlayerName = GameLib.GetPlayerUnit():GetName()
	tPacket.tPosition = self:buildVector3(tPosition.x, tPosition.y, tPosition.z)
		
	return tPacket
end

function MapPingMarker:onPacketRecieved(channel, tPacket, strSender)	
	if(self.tOptions.bEnableDebugMode) then
		self:DebugString("Packet Recieved")
		self:DebugTable(tPacket)
	end
				
	if(tPacket.strPlayerName == GameLib.GetPlayerUnit():GetName()) then
		self:setMarker(tPacket.tPosition)
	end
	
	for groupMemberIndex = 1, GroupLib.GetMemberCount(), 1 do		
		if(GroupLib.GetUnitForGroupMember(groupMemberIndex) ~= nil) then
			if(GroupLib.GetUnitForGroupMember(groupMemberIndex):GetName() == strSender) then
				self:setMarker(tPacket.tPosition)
			end
		end
	end
end

function MapPingMarker:sendPacket(tPacket)
	if(self.tOptions.bMarkOwnPings or self.tOptions.bEnableDebugMode) then
		if(self.tOptions.bEnableDebugMode) then
			self:DebugString("Sending Packet")
		end
		
		self:onPacketRecieved(self.comChannelName, tPacket, GameLib.GetPlayerUnit():GetName())
	end
	
	self.mComChannel:SendMessage(tPacket)
end


-----------------------------------------------------------------------------------------------
-- MapPingMarker Helper Functions
-----------------------------------------------------------------------------------------------

function MapPingMarker:buildVector3(xPos, yPos, zPos)
	local tPosition = {}
	tPosition.x = xPos
	tPosition.y = yPos
	tPosition.z = zPos
	
	return tPosition
end

function MapPingMarker:getDistanceTo(tPosition)
	tPlayerPos = GameLib.GetPlayerUnit():GetPosition()
	
	xd = tPosition.x - tPlayerPos.x
	yd = tPosition.y - tPlayerPos.y
	zd = tPosition.z - tPlayerPos.z
	
	return math.sqrt(math.pow(xd, 2) + math.pow(yd, 2) + math.pow(zd, 2))
end

function MapPingMarker:buildTriangle(tPositionA, tPositionB, tPositionC)
	local tTriangle = {}
	
	tTriangle.iSideA = self:getDistanceBetweenPoints(tPositionB, tPositionC)
	tTriangle.iSideB = self:getDistanceBetweenPoints(tPositionA, tPositionC)
	tTriangle.iSideC = self:getDistanceBetweenPoints(tPositionA, tPositionB)

	--Law of consines formulas, see http://mathworld.wolfram.com/LawofCosines.html
	tTriangle.iAngleA = math.acos(((-1 * math.pow(tTriangle.iSideA, 2)) + math.pow(tTriangle.iSideB, 2) + math.pow(tTriangle.iSideC, 2)) / (2 * tTriangle.iSideB * tTriangle.iSideC))
	tTriangle.iAngleB = math.acos((math.pow(tTriangle.iSideA, 2) + (-1 * math.pow(tTriangle.iSideB, 2)) + math.pow(tTriangle.iSideC, 2)) / (2 * tTriangle.iSideA * tTriangle.iSideC))
	tTriangle.iAngleC = math.acos((math.pow(tTriangle.iSideA, 2) + math.pow(tTriangle.iSideB, 2) + (-1 * math.pow(tTriangle.iSideC, 2))) / (2 * tTriangle.iSideA * tTriangle.iSideB))

	return tTriangle
end 

function MapPingMarker:getDistanceBetweenPoints(tPositionOne, tPositionTwo)
	local dX = tPositionTwo.x - tPositionOne.x
	local dY = tPositionTwo.y - tPositionOne.y
	
	return math.sqrt(math.pow(dX, 2) + math.pow(dY, 2))
end

function MapPingMarker:getUnitId(tUnit)
	if(tUnit ~= nil) then
		return tUnit:GetId()
	end	

	return 0
end

-- on SlashCommand "/mpm"
function MapPingMarker:ShowOptionsMenu()
	self.optionsMenu:Invoke() 
end

function MapPingMarker:setDistanceDisplayPositioning(left, top, right, bottom)
	self.tOptions.tDistanceWindowPositioning.left = left
	self.tOptions.tDistanceWindowPositioning.top = top 
	self.tOptions.tDistanceWindowPositioning.right = right 
	self.tOptions.tDistanceWindowPositioning.bottom = bottom 
end 

-----------------------------------------------------------------------------------------------
-- MapPingMarker OptionsMenu Functions
-----------------------------------------------------------------------------------------------

function MapPingMarker:onMarkOwnMapPings( wndHandler, wndControl, eMouseButton )
	self.tOptions.bMarkOwnPings = not self.tOptions.bMarkOwnPings
end

function MapPingMarker:onExit( wndHandler, wndControl, eMouseButton )
	self.optionsMenu:Close()
end

function MapPingMarker:onDisplayLock( wndHandler, wndControl, eMouseButton )
	self.distanceDisplay:SetStyle("Moveable", self.tOptions.bDistanceDisplayLocked)
	self.tOptions.bDistanceDisplayLocked = not self.tOptions.bDistanceDisplayLocked

end

function MapPingMarker:onEnableDebugMode( wndHandler, wndControl, eMouseButton )
	self.tOptions.bEnableDebugMode = not self.tOptions.bEnableDebugMode
end

function MapPingMarker:onProximityThresholdChance( wndHandler, wndControl, fNewValue, fOldValue )
	local flooredValue = math.floor(fNewValue)
	if(flooredValue == 1) then
		self.optionsMenu:FindChild("ProximityThreshold"):FindChild("ProxThresh Value"):SetText(tostring(flooredValue .. " Meter"))
	else
		self.optionsMenu:FindChild("ProximityThreshold"):FindChild("ProxThresh Value"):SetText(tostring(flooredValue .. " Meters"))
	end
	
	self.tOptions.iProximityThreshold = flooredValue
end

function MapPingMarker:onRestoreDefaults( wndHandler, wndControl, eMouseButton )
	self:restoreDefaults()
end

function MapPingMarker:onDistanceDisplayMove( wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom )
	self.distanceDisplay:GetAnchorOffsets()
end

function MapPingMarker:onEnableTrackMasterSupport( wndHandler, wndControl, eMouseButton )
	self.tOptions.bEnableTrackMasterSupport = not self.tOptions.bEnableTrackMasterSupport
end

-----------------------------------------------------------------------------------------
-- MapPingMarker Debug Functions
-----------------------------------------------------------------------------------------------

function MapPingMarker:DebugTable(tTable)
	for i, k in pairs(tTable) do
		if(type(k) == "table") then
			self:DebugTable(k)
		else 
			self:DebugString("Value " .. tostring(i) .. " = " .. tostring(k))
		end		
	end
end

function MapPingMarker:DebugString(strString)
	Print("[MapPingMaker]: " .. strString)
end

-----------------------------------------------------------------------------------------------
-- MapPingMarker Instance
-----------------------------------------------------------------------------------------------
local MapPingMarkerInst = MapPingMarker:new()
MapPingMarkerInst:Init()
